/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
     author: kekec 		E-mail: behaviour.brain.research@gmail.com
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          hst version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var item_cnt= 5;       // The number of stimuli in the task

var is_shuffle = true;   // Stimulus order: '1' = true = Randomly from the list of stimuli. '2' = false = Exactly according to the order of occurrence in the loaded list of stimuli

var is_lower = '3'; // Letter type: '1' = Original char set from text file, '2' = lowercase; '3' = UPPERCASE

var blank = '&#10062;';    // default placeholder for missing characters. see: setEmoji()

var is_sort = true; 	// Sort ascending chars (true || false)

var is_repeat = false; // Enable character selection multiple times: true || false

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var font_family = 'Arial, Helvetica, sans-serif';  // default canvas ctx.fontFamily 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item. (true || false)

var a_forward = false; // Automatically move to the next item: true || false

var keyboard= true;   // keyboard is enabled or disabled 

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 

var is_ltr = true;   	// Parsing (reading) direction: true = left-to-right, false = right-to-left see getScore() + stimGen()

var demo = false;       // if (demo == true) => Show task solutions 

var bclr_d = '#fff';  // Top line background color

var pclr_d = '#000';  // Font color - for top line

var bclr_s = '#e5e5e5'; 		// Bottom line background color				
 
var pclr_s = '#000';   // Font color - for bottom line

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var item;   				// current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  	// if  (dirty==true) => user changed any options in form 

var is_storage = false;		// local storage is enabled on user browser or not

var chars = {}; // default alphabetic chars object or loaded from txt file (user load character set from txt file) 

var stim = [];   // stimulus array.

var ans = [];    // user answer array. User's response to each item in the task

var moves = [];   // number of moves for each item. range: [0..item_cnt-1]

var chs = [];    // array of objects. For each item, remember which char the user selected (see: is_repeat)   

var is_def = 1;  // default stimuls and characters (1) or loaded from text file (0)

// array of default loaded sentences (stimuli, english proverbs). 30 sentences 
var sent = [
	'A BIRD IN HAND IS WORTH TWO IN THE BUSH',
	'THE PEN IS MIGHTIER THAN THE SWORD',
	'A CAT HAS NINE LIVES',
	'WHEN IN ROME DO AS THE ROMANS',
	'FORTUNE FAVORS THE BOLD',
	'BEST THINGS IN LIFE ARE FREE',
	'A DROWNING MAN WILL CLUTCH AT A STRAW',
	'AN EMPTY VESSEL MAKES MUCH NOISE',
	'THE SQUEAKY WHEEL GETS THE GREASE',
	'ALWAYS PUT YOUR BEST FOOT FORWARD',
	'A JOURNEY OF THOUSAND MILES BEGINS WITH A SINGLE STEP',
	'NO MAN IS AN ISLAND',
	'BARKING DOGS SELDOM BITE',
	'ALL GOOD THINGS COME TO AN END',
	'PEOPLE WHO LIVE IN GLASS HOUSES SHOULD NOT THROW STONES',
	'HOPE FOR THE BEST BUT PREPARE FOR THE WORST',
	'BETTER LATE THAN NEVER',
	'KEEP YOUR FRIENDS CLOSE AND YOUR ENEMIES CLOSER',
	'A PICTURE IS WORTH A THOUSAND WORDS',
	'THERE IS NO SUCH THING AS A FREE LUNCH',
	'NEVER LOOK A GIFT HORSE IN THE MOUTH',
	'ACTIONS SPEAK LOUDER THAN WORDS',
	'PRACTICE MAKES PERFECT',
	'TOO MANY COOKS SPOIL THE BROTH',
	'TWO WRONGS DO NOT MAKE A RIGHT',
	'ALL GOOD THINGS MUST COME TO AN END',
	'NECESSITY IS THE MOTHER OF INVENTION',
	'ZEAL WITHOUT KNOWLEDGE IS A RUNAWAY HORSE',
	'THE GRASS IS ALWAYS GREENER ON THE OTHER SIDE OF THE HILL',
	'NOTHING IS CERTAIN BUT DEATH AND TAXES'
	];



/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [create/start Timer, check storage, show start panel..]
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
	
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	
	document.addEventListener('keyup', onKeyUp);  //document.onkeyup = onKeyUp;
	
	loadDefault();
	
	loadFont();
	
	setEmoji();
	
	setIns();
	
	showStart();  

}


// load default sentences (stimuli) and characters set from default sent array
function loadDefault()
{
	let size= sent.length, key, txt; 
	for (let i = 0; i < size; i++)  {      
		txt = sent[i].replace(/\r\n|\n|\r/g, '');  
		txt= txt.trim();
		for (let j= 0, len = txt.length; j < len; j++) {
			key = txt[j]; 
			if (key.trim() != '')  {  // escape all whitespace key
				chars[key] = 1;
			}
		}				
	}		
	document.getElementById('sen_cnt').textContent= size;   // update default number of sentences (stimulus)  
	setItemCnt();
	setChar();	
}


// controls the keyboard during a task. 
function onKeyUp(e)  
{
	if (is_task && keyboard) 
	{
		e = e || window.e;   
		let c = e.key, c_low= c.toLocaleLowerCase(), c_up= c.toLocaleUpperCase(); //  c = e.key => character pressed
		if ( !is_repeat &&  ( chs[item][c] == 1  || chs[item][c_low] == 1 || chs[item][c_up] == 1 ) )  {  
			return;  // when character selection is allowed only once within each item
		}		
		if ( chars[c] == 1 || chars[c_low] == 1 || chars[c_up] == 1 )  {
			if ( is_lower == '2' )  {
				c= c.toLowerCase();
			}
			else if ( is_lower == '3' )  {
				c= c.toUpperCase(); 
			}	 //else { // not case. original character loaded from text file ( content of sent array )
			
			swap(c);			
		}
	}	
}
	

// set maximum item_cnt in task. check current item_cnt in task 
function setItemCnt()
{ 
	let e= document.getElementById('item_cnt');  // update maximum item_cnt (and current value for item_cnt)
	let sen_cnt= sent.length;
	if ( sen_cnt > 0 )  {
		e.min=  1;   
		e.max= sen_cnt;		
    //e.value = sen_cnt;
    let tmp=  parseInt(e.value, 10); // current item_cnt value		
		if ( tmp < 1 || tmp > sen_cnt )  {
			e.value= sen_cnt;  
		}
	}
	else  {
		e.min= 0;   
		e.max= 0;
    e.value= 0;		
	}
	document.getElementById('sen_cnt').textContent= sen_cnt;  // update Number of stimulus (sentences, sent) loaded 
}
 
 
// application timer handler function
function appTimer()  
{    
	if (timer_enabled && is_task)   // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd(); 
		}	
		else if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);	
		}
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 3; // 'Arial, Helvetica, sans-serif'
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[3];
}	


// fill Emoji Unicode Decimal values from array into select html element
function setEmoji()  
{
	// init array for emoji-s decimal values: https://www.quackit.com/character_sets/emoji/emoji_v3.0/unicode_emoji_v3.0_characters_all.cfm 
	let arr = [	'&#128512;', '&#128513;',	'&#128514;',	'&#129315;',	'&#128515;',	'&#128516;',	'&#128517;',	'&#128518;', '&#128521;',	'&#128522;',
						'&#8986;', '&#8987;', '&#9200;', '&#9748;', '&#9749;', '&#9875;', '&#9917;', '&#9924;', '&#9925;', '&#9940;', 
						'&#9962;', '&#9970;', '&#9973;', '&#9981;', '&#9986;', '&#9989;', '&#9992;', '&#9993;', '&#9995;', '&#10062;',
						'&#11088;', '&#127746;', '&#127749;', '&#127796;', '&#127797;', '&#127800;', '&#127802;', '&#127803;', '&#127805;', '&#127806;',
						'&#127807;', '&#127809;', '&#127812;', '&#127813;', '&#127815;', '&#127827;', '&#127925;', '&#127931;', '&#127984;', '&#128018;', 
						'&#128024;', '&#128054;', '&#128105;', '&#128107;', '&#128115;', '&#128116;', '&#128118;', '&#128123;', '&#128124;', '&#128152;' ]; 		
	const size = arr.length; // emoji-size
	// https://theasciicode.com.ar/extended-ascii-code/multiplication-sign-ascii-code-158.html
	arr.push('&times;');   
	arr.push('&copy;');   
	arr.push('&reg;');
	arr.push('_');
	let sel= document.getElementById('blank'), len = arr.length, opt, j= 1; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= i < size ? 'Emoji ' + j : 'Text ' + j;   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);
    if ( j == size ) {
			j= 1;
		}
		else  {
			j++;		
		}
	}                         
	sel.selectedIndex= 29; // '&#10062;'
	document.getElementById('blank_msg').innerHTML= sel.value; 
	blank= sel.value;
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('hst-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('hst-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		

  document.getElementById('ins_keyboard').style.display= keyboard ? 'inline' : 'none';	
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	

	document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	
	if ( document.getElementById('timer_enabled').checked )
  {   
    document.getElementById('ins_timer').setAttribute('style', 'display:block'); 		
		document.getElementById('ins_timer').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
		if (prg_type == '0')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will show how many items are left until the end of the task.';
		}
		else if (prg_type == '1')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
		}
		else  {
			document.getElementById('ins_timer').textContent+= ' But information about the remaining time will not be visible.';  
		}  
  }
	else  {
		document.getElementById('ins_timer').setAttribute('style', 'display:none');
	}	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
			if (e.validity.valid)  {
				msg.textContent= 'The maximum number of items is equal to the current number of loaded stimuli.'; 				
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
			} 
			break; 
    
		/*case 'is_shuffle':		
			msg.textContent= e.value == '1' ? '' : '';      
			break; */
		
		
		case 'is_lower':	
			switch ( e.value )	{
				case '1':
					msg.textContent= 'When using your own stimuli, probably the most reliable option is to choose the original. The text renders as it is. Before that, convert all letters to uppercase or lowercase in the uploaded file.';
					break;
				case '2':
					msg.textContent= 'Lowercase characters will always be displayed on the screen.';
					break;	
				case '3':
					msg.textContent= 'UPPERCASE characters will always be displayed on the screen.';
					break;		
      }     
			break; 
				
		case 'blank':
		  msg.innerHTML= e.value;
			break;
		
		case 'is_sort': 
			msg.textContent= e.checked ? 'Yes. We recommend using sorting only for left-to-right text direction.' : 'No.';  
			break; 
		
		case 'is_repeat': 
			msg.textContent= e.checked ? 'Yes. Within the same item, it will be possible to select the same character multiple times.' : 'No. When a character is selected for the first time within the current item, it will disappear from the screen.';  
			msg.textContent += ' Selecting this option significantly affects the size of the dependent variables in the task.';
			break;				
				
		case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No. Timer is disabled.';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
		
		/* case 'font_family':
		  break; */
			
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
		
		case 'a_forward':		
			msg.textContent= e.checked ? 'Yes. When the user successfully completes the current item, the application automatically moves on to the next task.' : 'No. To move on to the next task, the user must click on the button at the bottom right of the screen.';      
			break; 		
		
		case 'keyboard':		
			msg.textContent= e.checked ? 'Yes. The user can select letters on the screen by pressing the appropriate key on the keyboard. Drag and drop action can also be used.' : 'No. To select characters on the screen, the user must use only the drag and drop action.';      
			break; 
		
		case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.'; // The number of the current item will be displayed, as well as the total number of items.
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    

		case 'is_ltr':				  
			msg.textContent= e.value == true ? 'Default. Left-to-right text direction.' : 'Right-to-left text direction.';
			msg.textContent += ' Refers to loading stimuli from a file and parsing text in the top line.';
			break; 

		case 'demo': 
			msg.textContent= e.checked ? 'Yes. The content of the hidden sentence will be displayed at the bottom of the screen.' : 'No';  
			break; 
	
		case 'bclr_d': 
		case 'pclr_d': 
    case 'bclr_s': 
		case 'pclr_s': 
			msg.textContent= validColor(e.value)  ? '' : 'The selected color is invalid.'; 
			break; 	
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity most of options 
function optionsValid()  
{ 
  let res = document.getElementById('item_cnt').validity.valid 		 
				 && document.getElementById('timer').validity.valid 
				 && validColor(document.getElementById('bclr_d').value)
				 && validColor(document.getElementById('pclr_d').value)
				 && validColor(document.getElementById('bclr_s').value)
				 && validColor(document.getElementById('pclr_s').value); 
		
  return res;              
}


// start task (research) 
function startTask(e)   
{      			
  let ok= checkPar();
	
  if ( ok )	{		
	
		initLayout();
		
		setPar();
		
		stimGen();
		
		is_task= true;   // if ( timer_enabled ) finally "activate" timer 
		
		drawStim();    // draw first item (stimulus)
	}
	
}


// check (uploaded) stimulus (sentences) and characters
// return true if they are valid
function checkPar() 
{
	let sen_cnt= sent.length, chr_cnt= Object.keys(chars).length; 	
	let e = document.getElementById('is_valid'), msg= 'Unable to start task. ', ok = false; // validator message
		
	// To run a task you need to load stimuli (sentences, phrases) from a text file. The number of stimuli is zero or the number of loaded characters is zero	
	// Unable to start task. The number of currently loaded sentences (stimuli) is zero. Check the uploaded file.
	
	if ( sen_cnt == 0 || item_cnt == 0 ) {
		e.textContent= msg + 'The number of stimuli is zero. To run a task you need to load stimuli (sentences, phrases) from a text file.';
	}		
	else if ( sen_cnt < item_cnt ) {
		e.textContent= msg +	'The number of currently loaded sentences (' + sen_cnt + ') is less than the number of items (' + item_cnt + ').'; // The number of items is greater than the number of stimuli (sentences) loaded from the file.
	}
	else if ( chr_cnt < 3 || chr_cnt > 256 ) {
		e.textContent= msg + 'The number of characters loaded must be in the range: [3..256]. Check the uploaded file or reload the current document, ie press the reload button in your browser to reload the default settings.';  	
	}
	else  {
		ok= true;
		//e.textContent= '';
	}
		
	if ( !ok )  { 
    e.setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
	return ok;
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block';  	
	
 	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		//document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     //  items / remaining items
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // always reset  
	
	document.getElementById('demo_div').style.display= demo ? 'block' : 'none';   // Check button visible or not
	
  setChoices();
	
	/*
		bclr_d = '#fff';  // Top line background color
		pclr_d = '#000';  // Font color - for top line
		bclr_s = '#e5e5e5'; 		// Bottom line background color				 
		pclr_s = '#000';   // Font color - for bottom line
	*/
  document.getElementById('answer').style.backgroundColor= bclr_d;  // Top line background color
  document.getElementById('answer').style.color= pclr_d; 						// Font color - for top line
	document.getElementById('choice_cont').style.backgroundColor= bclr_s; 		// Bottom line background color	
	document.getElementById('choice_cont').style.color= pclr_s;   						// Font color - for bottom line
	
	document.getElementById('answer').style.fontFamily= font_family;
	document.getElementById('choice_cont').style.fontFamily= font_family;
}  


// set selection (choice) characters (bottom toolbar). These chars can be selected by the user (drag and drop or keyboard) and appear at the bottom row of the screen. 
function setChoices()
{
	let cont= document.getElementById('choice_cont');
	cont.innerHTML= '';
	let s_keys = Object.keys(chars);  // The Object.keys() method returns an array of a given object's own enumerable property names, iterated in the same order that a normal loop would. 
	if (is_sort) {
		s_keys.sort();  // sort array of object keys
	}	
	let len = s_keys.length, i= 1, val;  // val => drag and drop value (string visible on screen moved from bottom row to top row) 
	let fixed_c = getFixedChar(); 
	for (let j = 0; j < len; j++)  {
		let key = s_keys[j];  // sorted key
		let key_low= key.toLocaleLowerCase(), key_up= key.toLocaleUpperCase();
		if ( key.trim() == '' || fixed_c[key] || fixed_c[key_low] || fixed_c[key_up] ) {  
		  continue;   // escape fixed characters
		}
		let e = document.createElement('span'); //document.createElement('label');
    e.id = 'c_' + i;   // pattern for drag and drop nodes (choices in bottom row/menu) is id= 'c_1', 'c_2',..'c_n'
		if ( is_lower == '2' )  {
			val= key_low;    // key.toLowerCase();
		}
		else if ( is_lower == '3' )  {
			val= key_up;  // key.toUpperCase(); 
		}	
		else {
			val = key; // original character loaded from text file
		}		
		e.setAttribute('class', 'choice');
		e.textContent = val;
		//e.setAttribute('data-value', val);  
    e.title = val;
		e.setAttribute('draggable', 'true'); 
    e.setAttribute('ondragstart', 'dragStart(event)');   
		
    cont.appendChild(e);		
		cont.innerHTML += '&Tab;';  // or: '&#9;' or: '&#x9;' or: '&Tab;'
		i++;
	}

}


// set some internal application parameters (global vars)
function setPar()  
{  
	item = 0; 
	counter = 0;       // reset timer counter		
}


// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{		
	stim = [];	// reset stimulus array 
	ans = [];  	// reset user answer array 
	moves = []; // reset number of moves
	chs = [];   // reset selected chars by items
	
	let size= Object.keys(sent).length, obj= {}, ind, txt, j = 0; 
	let fixed_c = getFixedChar();
	for (let i= 0; i< item_cnt; i++)   
	{  // item_cnt => total items count in task (i = item)
		ans[i]= [];
		moves[i] = 0;   // reset number of moves
		chs[i] = {};    // 'empty' object for this ordinal item number => for selected chars
		do  
		{
			ind = is_shuffle ? Math.floor(Math.random() * size) : j; // array index
			txt= sent[ind];    //items[Math.floor(Math.random()*items.length)];
			txt = txt.trim();
			if ( !is_ltr ) {
				txt= reverseString(txt);
			}
			txt = txt.replace(/  +/g, ' '); // Regex to replace multiple spaces with a single space
			if ( is_lower == '2' )  {
				txt= txt.toLocaleLowerCase();   //txt.toLowerCase();
			}
			else if ( is_lower == '3' )  {
				txt= txt.toLocaleUpperCase();  //txt.toUpperCase(); 
			}				
    }	while ( obj[ind] == 1 );
		obj[ind] = 1;  // remember this choice (sentence)	
		stim.push(txt);
		// set default top row => answer array 
		for (let k= 0, len = txt.length; k < len; k++)	
		{
			let c = txt[k], c_low= c.toLocaleLowerCase(), c_up= c.toLocaleUpperCase();
			if ( c.trim() == '' || fixed_c [c] || fixed_c[c_low] || fixed_c[c_up] ) {  					
				ans[i].push(c);      // set fixed char or ' ' 
			}
			else {
				ans[i].push(blank);  // set default empty placeholder (Emoji)
			}
		}			
		j++;  // index for case if is_shuffle == false
	}	

}


// return all fixed characters (each character loaded from file with checked state)
function getFixedChar() {
	let col = document.getElementById('fixed_char').getElementsByClassName('char-fixed'), len = col.length, res = {};  // live HTMLCollection: class name for fixed chars in parent containter 
	for (let j= 0; j < len; j++) {    // col[j].checked must be true 
		res[ col[j].value ]= 1;
	} 
	return res;
}


 
// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// draw (repaint) current stimulus                     
function drawStim() 
{    		  
  // draw user anser (top part of screen)
	// https://css-tricks.com/almanac/properties/w/word-spacing/
	// https://css-tricks.com/almanac/properties/t/text-justify/
	document.getElementById('answer').innerHTML= '';
	for (let i= 0; i < ans[item].length; i++)  { 
		document.getElementById('answer').innerHTML+= ans[item][i]; 
	}

	// set characters visibility in bottom part of screen for curent item
	let col = document.getElementById('choice_cont').getElementsByClassName('choice'), len = col.length;  // live HTMLCollection: class name for fixed chars in parent containter 
	for (let i= 0; i < len; i++)  {    				
		let txt= col[i].textContent;   //  col[i].getAttribute('data-value');	
		col[i].style.visibility= ( is_repeat || chs[item][txt] != 1) ? 'visible' : 'hidden'; 	
	} 
	
	if (demo) {
		document.getElementById('demo_div').textContent= stim[item]; 
	}
	
	updateFooter();		
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// Changes application settings that can be changed directly by the user within the options form
// These variables below can be directly modified by the user on the options form
function setVarPar()  
{   
	
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  }
	
	is_shuffle= document.getElementById('is_shuffle').value == '2' ? false : true;   // Stimulus order: '1' = true = Randomly from the list of stimuli. '2' = false = Exactly according to the order of occurrence in the loaded list of stimuli
		
	is_lower= document.getElementById('is_lower').value; 
	
	blank = document.getElementById('blank').value; 
		
	is_sort= document.getElementById('is_sort').checked ? true : false;		
	
	is_repeat= document.getElementById('is_repeat').checked ? true : false;	
	
  timer_enabled= document.getElementById('timer_enabled').checked ? true : false;  // timer is enabled or disabled	
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
	font_family= document.getElementById('font_family').value;	
  
	is_back= document.getElementById('is_back').checked ? true : false; 
	
	a_forward= document.getElementById('a_forward').checked ? true : false;
	
	keyboard= document.getElementById('keyboard').checked ? true : false; 
	
	prg_type= document.getElementById('prg_type').value;  
		
	is_ltr= document.getElementById('is_ltr').value == '2' ? false : true;  // "2" => Right-to-Left. "1" => Left-to-Right (default)
	
  demo= document.getElementById('demo').checked ? true : false;	
	
	if (validColor(document.getElementById('bclr_d').value))  {		
		bclr_d= document.getElementById('bclr_d').value;
  }
	
	if (validColor(document.getElementById('pclr_d').value))  {		
		pclr_d= document.getElementById('pclr_d').value;
  }
	
	if (validColor(document.getElementById('bclr_s').value))  {		
		bclr_s= document.getElementById('bclr_s').value;
  }
	
	if (validColor(document.getElementById('pclr_s').value))  {		
		pclr_s= document.getElementById('pclr_s').value;
  }
	
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;
	
}


// test color for validity (from string)
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {    
		item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent=  (item + 1) + ' / ' + (item_cnt - item-1);  
	}
}


// ondragstart event for letters choices (html span objects in bottom row)
function dragStart(e)  {    
	// real MIME type "text/plain" being unsupported in IE11
	if ( is_task )  {  
		e.dataTransfer.setData('text', e.target.id );	 //  ev.dataTransfer.setData("text", ev.target.id);
	}
}


// ondragover() event for html element (top row) => <span id="answer" ondragover="allowDrop(event)" ondrop="drop(event)"></span>
// the ondragover event specifies wher the dragged data can be dropped
function allowDrop(e)  {
  if ( is_task )  {
		e.preventDefault();
	}
}


// ondrop() event for html element/object (top row)
// html: <span id="answer" ondragover="allowDrop(event)" ondrop="drop(event)"></span>
function drop(e)  
{  
  //	only works for this combination: "Down row" => "Up row" drag & drop 
  if ( is_task )
	{
		e.preventDefault();   
		
		let my_id = e.dataTransfer.getData('text'); // id.  Note: real MIME type "text/plain" being unsupported in IE11
		
		let c = document.getElementById(my_id).textContent;  // document.getElementById(my_id).getAttribute('data-value')
		
		swap(c);
	}
}


/* The final part of the drag and drop action (or keyboard selection). Sets the c character to the corresponding place in the stimulus (if that character exists within the stimulus). 
  c => character to change */
function swap(c)
{
	if ( c.trim() != '' )	
	{
		for (let i= 0, len= stim[item].length; i < len; i++)	{ 
			if (stim[item][i] == c || stim[item][i] == c.toLocaleLowerCase() || stim[item][i] == c.toLocaleUpperCase() )  {   
				ans[item][i]= c;				
			}
		}	
		if (!is_repeat)	{
			chs[item][c]= 1;  // only once within current item
		}
		moves[item]++;   // increase number of moves for current item
		
		if ( a_forward && getScore(item) == 1 ) {
			nextItem();
		}
		else  {
			drawStim();  // update/repaint stimulus
		}
	}
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// text html: https://www.howtoexcel.org/split-text-by-delimiter/
// Loads the user's sentences (and character set) from a text file
function loadStimuli() {
  let file = document.getElementById('file_dic').files[0];
	let reader = new FileReader();
	reader.addEventListener('load', function () 
	{	
		chars = {};
		sent = [];		
	  let lines = reader.result.split('\n'), size= lines.length;    //reader.result; // this is whole content from text file
		for (let i = 0; i < size; i++)  
		{      
			let txt = lines[i].toString().replace(/\r\n|\n|\r/g, '');  // redundant?
			txt= txt.trim();
			sent.push(txt);
			let len = txt.length, key;
			for (let j= 0; j < len; j++) {
        key = txt[j];  
				if ( key.trim() == '' || chars[key.toLocaleLowerCase()] == 1 || chars [key.toLocaleUpperCase()] == 1 )  {  					
					continue;  // escape all whitespace or if lowerCase or upperCase key exists
				}
				chars[key] = 1;
			}				
    }		
		let chr_cnt= Object.keys(chars).length;  // number of chars loaded from txt file
		//document.getElementById('sen_cnt').textContent= sent.length;   // update number of stimulus (sentences) loaded from txt file
		setItemCnt();
		setChar();
		is_def= 0;  
	}, false);
  
	/*reader.addEventListener('error', function() {
    alert('Error occurred reading file: ' + file.name);
  });*/

  if (file) {
    reader.readAsText(file);
  }
}


/* set default or user-loaded characters set 
   create in "Options" form set of chars checkbox where user can define is each char is fixed or not
	 chars keys are sorted ascending */
function setChar()
{
	
	let s_keys = Object.keys(chars);  // The Object.keys() method returns an array of a given object's own enumerable property names, iterated in the same order that a normal loop would. 
	s_keys.sort();  // sorted array of object keys
	
	let cont= document.getElementById('fixed_char');  // parent container
	cont.innerHTML= '';
	let i= 1;
	for (let j = 0, len = s_keys.length; j < len; j++) 
	{  			// j => index of sorted key object
		let key= s_keys[j];   // sorted char key
		let check = document.createElement('input');
    check.type = 'checkbox';
    check.id =   'char_' + i;
		check.value = key;
    check.title = key;
		
		check.setAttribute('onclick', 'toogleCheck(event)');				//check.onclick= toogleFixed(this);				
			//check.setAttribute('data-value', key);	
		
    let lab = document.createElement('label');
    lab.htmlFor = check.id;
    
    if ( key.trim() == '' )  {  // this should never happen
			check.checked= true;
			check.disabled= true;
			lab.appendChild(document.createTextNode('whitespace is a fixed character that you cannot select'));
			lab.title= 'Whitespace character';
		}
		else {
			lab.title= key;
			lab.appendChild(document.createTextNode(key));	
		}
		
    cont.appendChild(check);
    cont.appendChild(lab);
			
		let br = document.createElement('br');
    cont.appendChild(br);
		cont.innerHTML += '&Tab;';  // or: '&#9;' or: '&#x9;' or: '&Tab;'
		i++;
	}
	document.getElementById('chars_cnt').textContent= Object.keys(chars).length; // update number of characters loaded from file
}


// for each character loaded from file: toogle between fixed state (checked) or choice (unchecked)
function toogleCheck(e)  {
	let check = document.getElementById(e.target.id);
	if (check.checked) {
		check.setAttribute('class', 'char-fixed');
	}
	else  {
		check.removeAttribute('class');
	}
}


// https://www.freecodecamp.org/news/how-to-reverse-a-string-in-javascript-in-3-different-ways-75e4763c68cb/
function reverseString(str) {
  return str.split("").reverse().join("");
}


// return number of points for item i
function getScore(i) 
{
	/*if ( !is_ltr )  {
		ans[i].reverse();  // stim[i] is reversed in stimGen()  
	}*/
	
	let answer= ans[i].join('');  // returns a string with all array elements joined 
	
	if ( answer.length == 0 || stim[i].length == 0 )  {
		return 0;
	}
  else if ( answer == stim[i] || answer.toLocaleLowerCase() == stim[i].toLocaleLowerCase() || answer.toLocaleUpperCase() == stim[i].toLocaleUpperCase() ) {
		return 1;
	} 
	else  {
		return 0;
	}	
}


/* return number of different characters within item i
	or minimum possible number of moves for this item 
	par: i == item number */
function getMin(i) 
{
	let obj = {}, res= 0;
	let fixed_c = getFixedChar(); 
	for ( let j= 0, len= stim[i].length; j < len; j++ )	 
	{
		let c = stim[i][j], c_low= c.toLocaleLowerCase(), c_up= c.toLocaleUpperCase();
		if ( c.trim() == '' || fixed_c [c] || fixed_c[c_low] || fixed_c[c_up] ) {
		  continue;  // escape whitespace + all fixed characters
		}
		if ( typeof  obj [ stim[i][j] ] === 'undefined' )  {
			obj [ stim[i][j] ] = 1;
			res++;
		}
	}
	return res;
}


// end of task (research)
function theEnd()  
{    
	is_task = false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  //document.getElementById('answer').style.display= 'none';  
	document.getElementById('result_wrapper').style.display= 'block';
  
	const D = 3;  // number of decimals
	let header= [ 'Item',  'Stimulus', 'Answer', 'Points', 'Moves', 'Min-moves', 'Redundancy' ]; // 7-columns
	let size= stim.length,  score = 0,  arr = [], rows = [],  sum_min = 0, sum_moves = 0, mis = 0;		
	for (let i = 0; i < size; i++)  		// for (let i = 0; i < item_cnt; i++)	
	{				
		let p = getScore(i);   // number of points in this i-item			
		score+= p;  // sum of points in whole task
		let txt = ans[i].join('');   // user answer to i-item
		sum_moves+= moves[i];
		let min_m = getMin(i);  // maximum possible number of moves for this item
		sum_min+= min_m;		
		if ( p > 0 ) {  
		  // compute redundancy. Always is : (moves[i] > min_m) => ( moves[i] / min_m - 1) >= 0 => redundancy is positive (>= 0) 
			let redun =  moves[i] / min_m - 1;   //  redundancy should be > 0 => moves[i] / min_m >= 1 => redundancy is always >= 0.			
			arr.push(redun);  // redundancy only for completed items are push for descriiptive statistics
			rows[i]= [ parseInt(i+1,10), stim[i], txt,  p,  moves[i], min_m,  redun.toFixed(D) ];
		}
		else {
			mis++;			
			rows[i]= [ parseInt(i+1,10), stim[i], txt,  p,  moves[i], min_m,  '' ];
		}
	}  
	let n = arr.length;  // number of non-missing items (for redundancy var)
	let m = n > 0  ?  mean(arr).toFixed(D) : '';      
	           // [ 'Item',   'St', 'An',    'Points',                 'Moves',                    'Min-moves',           'Redundancy' ]; // 7-columns
	rows.push (  [ 'Average', '',    '',   (score/size).toFixed(D),  (sum_moves/size).toFixed(D), (sum_min/size).toFixed(D),  m ] );   // totals (sums)   7-columns	
  createTableArray(document.getElementById('table1'), rows, header); 
  	
	// table2 => Descriptive statistics	
	rows = [];  // reset table rows (for table2)
	// n and mean are calculated before
	let sd =  n > 1  ?  stDev(arr, m).toFixed(D) : '';   	 	
  rows.push( [ m, sd, n ] );  // last row
	
	header= ['Average', 'Standard deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 

  // update 'score' label => above "Repeat" button	
	document.getElementById('score').textContent= 'Sum of points= ' + score + '. '; 
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent+= 'Descriptive statistics cannot be calculated. ';
			break;
		case 1: 
			document.getElementById('score').textContent+= 'Average Redundancy= ' + m + '  Standard deviation cannot be calculated. '; 
			break;	
		default:
			document.getElementById('score').textContent+= 'Average Redundancy= ' + m + '  Standard Deviation= ' + sd + '. ';
	}
	document.getElementById('score').textContent+= 'Non-completed items= ' + mis;
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) { 
			setStorage(score, mis, m, sd);   			
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  

}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if (dd < 10) { dd='0'+dd; } 
	if (mm < 10) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 score = sum (number) of points in task
	 incomplete = number of incomplete items in task
   mean = average (mean) for Redundancy 
   sd = standard deviation for Redundancy	 */   	 
function setStorage(score, incomplete, mean, sd) 
{    
	let chr_cnt= Object.keys(chars).length, fix_cnt = document.getElementById('fixed_char').getElementsByClassName('char-fixed').length;

	let oldItems = JSON.parse(localStorage.getItem('hst-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   			
			'score' : score,
			'incomplete' : incomplete,
			'mean'  : mean, 	
			'sd' : sd,								
			'default': is_def,			
			'char-total' : chr_cnt,
			'char-fixed': fix_cnt,			
			'items' : item_cnt,
			'order' : is_shuffle ? 1 : 2,
			'letter-type' : is_lower,			
			'sort' : is_sort ? 1 : 0,
			'multiple' : is_repeat ? 1 : 0,			
			'timer' : timer_enabled ? timer : '',			
			'is-back' : is_back ? 1 : 0,
			'auto-move' : a_forward ? 1 : 0,			
			'keyboard' : keyboard ? 1 : 0,
			'parse-dir' : is_ltr ? 1 : 2, 			
			'demo' : demo ? 1 : 0
  };

  oldItems.push(newItem);

  localStorage.setItem('hst-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('hst-v1.0')  ?  JSON.parse(localStorage.getItem('hst-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('hst-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('hst-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('hst-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
